import numpy as np
import pandas as pd
from tqdm import tqdm
from joblib import Parallel, delayed

def apply_df(input_args):
    df, func, kwargs = input_args
    if "progress_bar" in kwargs:
        progress_bar = kwargs.pop('progress_bar')
    else:
        progress_bar = False
    if "args" in kwargs:
        args_ = kwargs.pop('args')
    else:
        args_ = None
    if progress_bar:
        tqdm.pandas(leave=False, desc=func.__name__, unit='emails', dynamic_ncols=True, mininterval=2.0)
        df = df.progress_apply(func, axis=1, args=args_)
    else:
        df = df.apply(func, axis=1, args=args_)
    return df

def apply_by_multiprocessing(df, func, **kwargs):
    """Apply a function along an axis of the DataFrame using multiprocessing.
    A maximum of half of the core available in the system will be used.

    Parameters
    ----------
    df : pd.DataFrame
        DataFrame where the function is applied

    func : function to apply

    Returns
    -------
    pd.DataFrame
        Returns the DataFrame with the function applied.
    """
    workers = kwargs.pop('workers')
    workers = min(workers, int(df.shape[0]/2))
    workers = max(workers, 1)
    if (df.shape[0]==1) or (workers==1):
        return apply_df((df, func, kwargs))
    retLst = Parallel(n_jobs=workers)(delayed(apply_df)(input_args=(d, func, kwargs)) for d in np.array_split(df, workers))
    return pd.concat(retLst)



import logging
from gensim.models import Word2Vec
from gensim.models.keyedvectors import Vocab, KeyedVectors
from sklearn.feature_extraction.text import TfidfVectorizer
from sklearn.decomposition import TruncatedSVD
from sklearn.feature_extraction.text import CountVectorizer
# from glove import Corpus, Glove

from melusine.utils.streamer import Streamer

log = logging.getLogger('Embeddings')
log.setLevel(logging.DEBUG)
formatter = logging.Formatter('%(asctime)s - %(name)s - %(levelname)s - %(message)s',
                              datefmt='%d/%m %I:%M')


class Embedding:
    """Class to train embeddings with Word2Vec algorithm.
    Attributes
    ----------
    input_column : str,
        String of the input column name for the pandas dataframe to compute the embedding on.
    streamer : Streamer instance,
        Builds a stream a tokens from a pd.Dataframe to train the embeddings.
    word2id: dict,
        Word vocabulary (key: word, value: word_index.
    embedding : Gensim KeyedVector Instance,
        Gensim KeyedVector Instance relative to the specific trained or imported embedding.
    method : str,
        One of the following :
            - "word2vec_sg" : Trains a Word2Vec Embedding using the Skip-Gram method (usually takes a long time).
            - "word2vec_ns" : Trains a Word2Vec Embedding using the Negative-Sampling method.
            - "word2vec_cbow" : Trains a Word2Vec Embedding using the Continuous Bag-Of-Words method.
            - "lsa_docterm" : Trains an Embedding by using an SVD on a Document-Term Matrix.
            - "lsa_tfidf" : Trains an Embedding by using an SVD on a TF-IDFized Document-Term Matrix.
            - "glove" : Trains a GloVe Embedding. NOT IMPLEMENTED YET.
    train_params : dict,
        Additional parameters for the embedding training. Check the following documentation :
            - gensim.models.Word2Vec for Word2Vec Embeddings
            - sklearn.decomposition.TruncatedSVD for LSA Embeddings
            - glove.Glove for GloVe Embeddings
        If left untouched, the default training values will be kept from the aforementioned packages.
    Examples
    --------
    >>> from melusine.nlp_tools.embedding import Embedding
    >>> embedding = Embedding()
    >>> embedding.train(X)
    >>> embedding.save(filepath)
    >>> embedding = Embedding().load(filepath)
    """

    def __init__(self,
                 input_column='clean_text',
                 workers=40,
                 random_seed=42,
                 iter=15,
                 size=300,
                 method="word2vec_cbow",
                 stop_removal=True,
                 min_count=100
                 ):
        """
        Parameters :
        ----------
        input_column : str,
            String of the input column name for the pandas dataframe to compute the embedding on (default="clean_text").
        workers : int,
            Number of CPUs to use for the embedding training process (default=40).
        random_seed : int,
            Seed for reproducibility (default=42).
        iter : int,
            Number of epochs (default=15). Used for Word2Vec and GloVe only.
        size : int,
            Desired embedding size (default=300).
        window : int,
            If Word2Vec, window used to find center-context word relationships.
            If GloVe, window used to compute the co-occurence matrix.
        min_count : int,
            Minimum number of appeareance of a token in the corpus for it to be kept in the vocabulary (default=100).
        stop_removal : bool,
            If True, removes stopwords in the Streamer process (default=True).
        method : str,
            One of the following :
                - "word2vec_sg" : Trains a Word2Vec Embedding using the Skip-Gram method (usually takes a long time).
                - "word2vec_ns" : Trains a Word2Vec Embedding using the Negative-Sampling method.
                - "word2vec_cbow" : Trains a Word2Vec Embedding using the Continuous Bag-Of-Words method.
                - "lsa_docterm" : Trains an Embedding by using an SVD on a Document-Term Matrix.
                - "lsa_tfidf" : Trains an Embedding by using an SVD on a TF-IDFized Document-Term Matrix.
                - "glove" : Trains a GloVe Embedding. NOT IMPLEMENTED YET.
        min_count : TODO
        """

        self.logger = logging.getLogger('NLUtils.Embedding')
        ch = logging.StreamHandler()
        ch.setLevel(logging.INFO)
        ch.setFormatter(formatter)
        self.logger.addHandler(ch)
        self.logger.debug('Create an Embedding instance.')
        self.input_column = input_column
        self.streamer = Streamer(column=self.input_column, stop_removal=stop_removal)
        self.word2id = {}
        self.embedding = None
        self.method=method

        if self.method in ['word2vec_sg', 'word2vec_cbow', 'word2vec_ns']:
            self.train_params = {
                 "size":size,
                 "alpha":0.025,
                 "window":5,
                 "min_count":min_count,
                 "max_vocab_size":None,
                 "sample":0.001,
                 "seed":random_seed,
                 "workers":3,
                 "min_alpha":0.0001,
                 "ns_exponent":0.75,
                 "cbow_mean":1,
                 "iter":iter,
                 "null_word":0,
                 "trim_rule":None,
                 "sorted_vocab":1,
                 "batch_words":10000,
                 "compute_loss":False,
                 "callbacks":(),
                 "max_final_vocab":None
                 }
        elif self.method in 'lsa_tfidf':
            self.train_params={
            #TfidfVectorizer Parameters
                "vectorizer_encoding":'utf-8',
                "vectorizer_decode_error":'strict',
                "vectorizer_strip_accents":None,
                "vectorizer_lowercase":False,
                "vectorizer_analyzer":'word',
                "vectorizer_stop_words":None,
                "vectorizer_ngram_range":(1, 1),
                "vectorizer_max_df":1.0,
                "vectorizer_min_df":1,
                "vectorizer_max_features":None,
                "vectorizer_vocabulary":None,
                "vectorizer_binary":False,
                "vectorizer_norm":'l2',
                "vectorizer_use_idf":True,
                "vectorizer_smooth_idf":True,
                "vectorizer_sublinear_tf":False,

            #TruncatedSVD Parameters
                "svd_n_components":size,
                "svd_algorithm":'randomized',
                "svd_n_iter":iter,
                "svd_random_state":random_seed,
                "svd_tol":0.0
            }
        elif self.method == 'lsa_docterm':
            self.train_params = {
            #CountVectorizer Parameters
                "vectorizer_encoding":'utf-8',
                "vectorizer_decode_error":'strict',
                "vectorizer_strip_accents":None,
                "vectorizer_lowercase":False,
                "vectorizer_stop_words":None,
                "vectorizer_ngram_range":(1, 1),
                "vectorizer_analyzer":'word',
                "vectorizer_max_df":1.0,
                "vectorizer_min_df":1,
                "vectorizer_max_features":None,
                "vectorizer_vocabulary":None,
                "vectorizer_binary":False,

            #TruncatedSVD Parameters
                "svd_n_components":size,
                "svd_algorithm":'randomized',
                "svd_n_iter":iter,
                "svd_random_state":random_seed,
                "svd_tol":0.0
                 }
        else:
            raise ValueError(f"Error: Embedding method {method} not recognized or not implemented yet ;).")

    def save(self, filepath):
        """Method to save Embedding object."""
        self.embedding.save(filepath)

    def load(self, filepath):
        """Method to load Embedding object."""
        self.embedding = KeyedVectors.load(filepath, mmap='r')
        return self

    def train(self, X):
        """Train embeddings with the desired word embedding algorithm (default is Word2Vec).
        Parameters
        ----------
        X : pd.Dataframe
            Containing a clean body column.
        embedding_type: str
            Desired embedding type
        """

        self.logger.info('Start training for embedding')

        self.streamer.to_stream(X)

        if self.method in ['word2vec_sg','word2vec_ns', 'word2vec_cbow']:
            self.train_word2vec()
        elif self.method == 'lsa_tfidf':
            self.train_tfidf()
        elif self.method == 'lsa_docterm':
            self.train_docterm()
        #elif embedding_type == 'glove':
        #    self.train_glove(X, self.input_column)

        self.logger.info('Done.')

    def train_tfidf(self):
        """Train a TF-IDF Vectorizer to compute a TF-IDFized Doc-Term Matrix relative to a corpus.
        Parameters
        ----------
        X : pd.Dataframe
            Containing a column with tokeized documents.
        input_column: str
            Name of the input column containing tokens
        """

        def dummy_function(doc):
            return doc

        tfidf_vec = TfidfVectorizer(
            tokenizer=dummy_function,
            preprocessor=dummy_function,
            token_pattern=None,

            encoding=self.train_params["vectorizer_encoding"],
            decode_error=self.train_params["vectorizer_decode_error"],
            strip_accents=self.train_params["vectorizer_strip_accents"],
            lowercase=self.train_params["vectorizer_lowercase"],
            analyzer=self.train_params["vectorizer_analyzer"],
            stop_words=self.train_params["vectorizer_stop_words"],
            ngram_range=self.train_params["vectorizer_ngram_range"],
            max_df=self.train_params["vectorizer_max_df"],
            min_df=self.train_params["vectorizer_min_df"],
            max_features=self.train_params["vectorizer_max_features"],
            vocabulary=self.train_params["vectorizer_vocabulary"],
            binary=self.train_params["vectorizer_binary"],
            use_idf=self.train_params["vectorizer_use_idf"],
            smooth_idf=self.train_params["vectorizer_smooth_idf"],
            sublinear_tf=self.train_params["vectorizer_sublinear_tf"]
        )

        tfidf_data = tfidf_vec.fit_transform(self.streamer.stream)

        self.word2id = tfidf_vec.vocabulary_
        embedding_matrix = self.train_svd(tfidf_data)

        self.create_keyedvector_from_matrix(embedding_matrix, self.word2id)

    def train_docterm(self):
        """Train a Count Vectorizer to compute a Doc-Term Matrix relative to a corpus.
        Parameters
        ----------
        X : pd.Dataframe
            Containing a column with tokenized documents
        input_column: str
            Name of the input column containing tokens
        """

        def dummy_function(doc):
            return doc

        # CountVectorizer

        count_vec = CountVectorizer(
            tokenizer=dummy_function,
            preprocessor=dummy_function,
            token_pattern=None,

            encoding=self.train_params["vectorizer_encoding"],
            decode_error=self.train_params["vectorizer_decode_error"],
            strip_accents=self.train_params["vectorizer_strip_accents"],
            lowercase=self.train_params["vectorizer_lowercase"],
            stop_words=self.train_params["vectorizer_stop_words"],
            ngram_range=self.train_params["vectorizer_ngram_range"],
            analyzer=self.train_params["vectorizer_analyzer"],
            max_df=self.train_params["vectorizer_max_df"],
            min_df=self.train_params["vectorizer_min_df"],
            max_features=self.train_params["vectorizer_max_features"],
            vocabulary=self.train_params["vectorizer_vocabulary"],
            binary=self.train_params["vectorizer_binary"]
        )

        docterm_data = count_vec.fit_transform(self.streamer.stream)

        self.word2id = count_vec.vocabulary_
        embedding_matrix = self.train_svd(docterm_data)

        self.create_keyedvector_from_matrix(embedding_matrix, self.word2id)

    def train_svd(self, vectorized_corpus_data):
        """Fits a TruncatedSVD on a Doc-Term/TF-IDFized Doc-Term Matrix for dimensionality reduction.
        Parameters
        ----------
        vectorized_corpus_data: CountVectorizer or TfidfVectorizer object,
            Sklearn object on which the TruncatedSVD will be computed.
        """

        svd=TruncatedSVD(n_components=self.train_params["svd_n_components"],
                        algorithm=self.train_params["svd_algorithm"],
                        n_iter=self.train_params["svd_n_iter"],
                        random_state=self.train_params["svd_random_state"],
                        tol=self.train_params["svd_tol"])
        svd.fit(vectorized_corpus_data)

        embedding_matrix = svd.components_.T
        return embedding_matrix

    def train_word2vec(self):
        """Fits a Word2Vec Embedding on the given documents, and update the embedding attribute.
        """

        if self.method== "word2vec_sg":
            self.train_params["negative"]=1
            self.train_params["sg"]=1
            self.train_params["hs"]=1
        elif self.method == "word2vec_ns" :
            self.train_params["negative"]=1
            self.train_params["sg"]=1
            self.train_params["hs"]=0
        elif self.method == "word2vec_cbow" :
            self.train_params["negative"]=5
            self.train_params["sg"]=0
            self.train_params["hs"]=0

        embedding = Word2Vec(size=self.train_params["size"],
                            alpha=self.train_params["alpha"],
                            window=self.train_params["window"],
                            min_count=self.train_params["min_count"],
                            max_vocab_size=self.train_params["max_vocab_size"],
                            sample=self.train_params["sample"],
                            seed=self.train_params["seed"],
                            workers=self.train_params["workers"],
                            min_alpha=self.train_params["min_alpha"],
                            sg=self.train_params["sg"],
                            hs=self.train_params["hs"],
                            negative=self.train_params["negative"],
                            ns_exponent=self.train_params["ns_exponent"],
                            cbow_mean=self.train_params["cbow_mean"],
                            iter=self.train_params["iter"],
                            null_word=self.train_params["null_word"],
                            trim_rule=self.train_params["trim_rule"],
                            sorted_vocab=self.train_params["sorted_vocab"],
                            batch_words=self.train_params["batch_words"],
                            compute_loss=self.train_params["compute_loss"],
                            callbacks=self.train_params["callbacks"],
                            max_final_vocab=self.train_params["max_final_vocab"])

        #TODO Fix Streamer
        embedding.build_vocab(list(self.streamer.stream))
        embedding.train(list(self.streamer.stream),
                        total_examples=embedding.corpus_count,
                        epochs=self.train_params["iter"])

        self.embedding = embedding.wv



    #def train_glove(self, X, input_column):
    #
    #    corpus = Corpus()
    #
    #    corpus.fit(X[input_column], window=self.window)
    #
    #    self.word2id = corpus.dictionary
    #
    #    glove=Glove(no_components=self.size)
    #    glove.fit(corpus.matrix, epochs=self.iter, no_threads=self.workers, )
    #
    #    self.create_keyedvector_from_matrix(glove.word_vectors, self.word2id)

    def create_keyedvector_from_matrix(self, embedding_matrix, word2id):
        """
        Imports the necessary attributes for the Embedding object from an embedding matrix and a word2id vocabulary. Can be used for custom pre-trained embeddings.
        Parameters
        ----------
        embedding_matrix: numpy.ndarray
            Embedding matrix as a numpy object
        word2id: dict
            Word vocabulary (key: word, value: word_index)
        """

        vocab = {word:word2id[word] for word in sorted(word2id, key=word2id.__getitem__, reverse=False)}
        embedding_matrix = embedding_matrix
        vector_size = embedding_matrix.shape[1]

        kv = KeyedVectors(vector_size)
        kv.vector_size = vector_size
        kv.vectors = embedding_matrix

        kv.index2word = list(vocab.keys())

        kv.vocab = {word: Vocab(index=word_id, count=0) for word, word_id in vocab.items()}

        self.embedding = kv



import numpy as np
import pandas as pd
import scipy.sparse as sp
from collections import Counter
from sklearn.base import BaseEstimator, TransformerMixin
from sklearn.feature_extraction.text import TfidfVectorizer
from melusine.utils.multiprocessing import apply_by_multiprocessing
from melusine.config.config import ConfigJsonReader

conf_reader = ConfigJsonReader()
config = conf_reader.get_config_file()

keywords = config["words_list"]["keywords"]
stopwords = config["words_list"]["stopwords"] + config["words_list"]["names"]


class KeywordsGenerator(BaseEstimator, TransformerMixin):
    """Class to extract list of keywords from text.

    It is compatible with scikit-learn API (i.e. contains fit, transform
    methods).

    Parameters
    ----------

    max_tfidf_features : int, optional
        Size of vocabulary for tfidf.
        Default value, 10000.

    keywords : list, optional
        Keywords to extracted as priority.
        Default value, "keywords" list defined in conf file.

    stopwords : list, optional
        Stopwords not to be extracted.
        Default value, "names" and "stopwords" lists defined in conf file.

    resample : bool, optional
        True if dataset must be resampled according to class distribution,
        else False.
        Default value, True.

    n_jobs : int, optional
        Number of cores used for computation.
        Default value, 20.

    copy : bool, optional
        Make a copy of DataFrame.
        Default value, True.

    n_max_keywords : int, optional
        Maximum number of keywords to be returned.
        Default value, 6.

    n_min_keywords : int, optional
        Minimum number of keywords to be returned.
        Default value, 0.

    threshold_keywords : float, optional
        Minimum tf-idf score for word to be selected as keyword.
        Default value, 0.0.

    n_docs_in_class : int, optional
        Number of documents in each classes.
        Default value, 100.

    keywords_coef : int, optional
        Coefficient multiplied with the tf-idf scores of each keywords.
        Default value, 10.

    Attributes
    ----------
    max_tfidf_features, keywords, stopwords, resample, n_jobs, progress_bar,
    copy, n_max_keywords, n_min_keywords, threshold_keywords, n_docs_in_class,
    keywords_coef,

    tfidf_vectorizer : TfidfVectorizer instance from sklearn,

    dict_scores_ : dictionary,
        Tf-idf scores for each tokens.

    max_score_ : np.array,

    Examples
    --------
    >>> from melusine.summarizer.keywords_generator import KeywordsGenerator
    >>> keywords_generator = KeywordsGenerator()
    >>> keywords_generator.fit(X, y)
    >>> keywords_generator.transform(X)
    >>> print(X['keywords'])

    """

    def __init__(self,
                 max_tfidf_features=10000,
                 keywords=keywords,
                 stopwords=stopwords,
                 resample=False,
                 n_jobs=20,
                 progress_bar=True,
                 copy=True,
                 n_max_keywords=6,
                 n_min_keywords=0,
                 threshold_keywords=0.0,
                 n_docs_in_class=100,
                 keywords_coef=10):
        self.max_tfidf_features_ = max_tfidf_features
        self.tfidf_vectorizer = TfidfVectorizer(max_features=max_tfidf_features
                                                )
        self.keywords = keywords
        self.stopwords = stopwords
        self.resample = resample
        self.n_jobs = n_jobs
        self.progress_bar = progress_bar
        self.copy = copy
        self.n_max_keywords = n_max_keywords
        self.n_min_keywords = n_min_keywords
        self.threshold_keywords = threshold_keywords
        self.n_docs_in_class = n_docs_in_class
        self.keywords_coef = keywords_coef

    def fit(self, X, y=None):
        """Fit the weighted tf-idf model with input data.

        If resample attribute is True the dataset will be resampled according
        to class distribution.

        Parameters
        ----------
        X : pandas.DataFrame, shape (n_samples, n_features)
            X must contain ['tokens'] column.

        y : Ignored

        Returns
        -------
        self : object
            Returns the instance itself.
        """
        if self.resample:
            X_resample = self.resample_docs(X, y)
        else:
            X_resample = X

        X_resample['tokens'] = X_resample['tokens'].apply(self._remove_stopwords)

        # fit tf-idf on resample data set
        tokens_joined = X_resample['tokens'].apply(lambda x: ' '.join(x))
        self.tfidf_vectorizer.fit(tokens_joined)

        # modify the idf weights given frequency in the corpus
        idf_weights = self._add_tf_to_idf(X_resample)
        self.tfidf_vectorizer._tfidf._idf_diag = sp.spdiags(idf_weights,
                                                            diags=0,
                                                            m=len(idf_weights),
                                                            n=len(idf_weights))

        # return vetorizer with binary term frequency atribute
        self.dict_scores_ = dict(zip(self.tfidf_vectorizer.get_feature_names(),
                                     self.tfidf_vectorizer.idf_))
        self.max_score_ = np.max(self.tfidf_vectorizer.idf_)

        return self

    def transform(self, X):
        """Returns list of keywords in apparition order for each document
        with the weighted tf-idf already fitted.

        Parameters
        ----------
        X : pandas.DataFrame, shape (n_samples, n_features)
            X must contain ['tokens'] column.

        Returns
        -------
        X_new : pandas.DataFrame, shape (n_samples, n_components)
        """
        if self.copy:
            X_ = X.copy()
        else:
            X_ = X

        X_['keywords'] = apply_by_multiprocessing(df=X_[['tokens']],
                                                  func=self.get_keywords,
                                                  axis=1,
                                                  workers=self.n_jobs,
                                                  progress_bar=self.progress_bar)

        return X_

    def get_keywords(self, row):
        """Returns list of keywords in apparition order with the
        weighted tf-idf already fitted.

        Parameters
        ----------
        row : row of pd.Dataframe, columns ['tokens']

        Returns
        -------
        list of strings
        """
        tokens = self._remove_stopwords(row['tokens'])
        tokens = [x for x in tokens if not x.isdigit()]
        scores = Counter({t: self.dict_scores_.get(t, 0) for t in tokens})
        n = sum(i > self.threshold_keywords for i in list(scores.values()))
        n = min(n, self.n_max_keywords)
        n = max(n, self.n_min_keywords)
        keywords = [x[0] for x in scores.most_common(n)]
        index_sorted = [(k, tokens.index(k)) for k in keywords if k in tokens]
        index_sorted = sorted(index_sorted, key=lambda x: x[1])
        keywords_sorted = [i[0] for i in index_sorted]

        return keywords_sorted

    def resample_docs(self, X, y=None):
        """Method for resampling documents according to class distribution."""
        X_ = X.copy()
        if y is not None:
            X_['label'] = y
        X_['split'] = 0
        for c in X_.label.unique():
            N_c = X_[X_["label"] == c].shape[0]
            I_c = np.random.randint(0, self.n_docs_in_class+1, N_c)
            X_.loc[X_["label"] == c, 'split'] = I_c

        X_resample = pd.DataFrame(
            X_[['label', 'split', 'tokens']]
            .groupby(['label', 'split'], as_index=False)['tokens']
            .sum()
        )

        return X_resample

    def _remove_stopwords(self, tokens):
        """Method to filter stopwords from potential list of keywords."""
        return [t for t in tokens if t not in self.stopwords]

    def _add_tf_to_idf(self, X):
        """Returns the tf-idf weights of each tokens"""
        tokens_joined = X['tokens'].apply(lambda x: ' '.join(x))
        X_vec = self.tfidf_vectorizer.transform(tokens_joined)
        feature_names = self.tfidf_vectorizer.get_feature_names()
        idf_weights = self._get_weights(X_vec.toarray(),
                                        self.keywords,
                                        feature_names)

        return idf_weights

    def _get_weights(self, X_vec, keywords_list, feature_names):
        """Put max weights for each word of redistributed mails."""
        max_ = np.max(X_vec, axis=0)
        mmax_ = np.max(max_)
        for k in keywords_list:
            if k in feature_names:
                max_[feature_names.index(k)] = mmax_ * self.keywords_coef

        return max_



import re
from melusine.config import ConfigJsonReader

conf_reader = ConfigJsonReader()
config = conf_reader.get_config_file()
regex_transition_list = config['regex']['build_historic']['transition_list']


def build_historic(row):
    """Rebuilds and structures historic of emails from the whole contents.
    Function has to be applied with `apply` method of a DataFrame along an
    axis=1.
    For each email of the historic, it segments the body into 2 different parts
    (2 keys of dict):

    {'text': extract raw text without metadata,
     'meta': get transition from the 'transition_list' defined in the conf.json
     }.


    Parameters
    ----------
    row : row,
        A pandas.DataFrame row object with 'body' column.

    Returns
    -------
    list

    Examples
    --------
        >>> import pandas as pd
        >>> data = pd.read_pickle('./tutorial/data/emails_anonymized.pickle')
        >>> # data contains a 'body' column

        >>> from melusine.prepare_email.build_historic import build_historic
        >>> build_historic(data.iloc[0])  # apply for 1 sample
        >>> data.apply(build_historic, axis=1)  # apply to all samples

    """
    email_body = row['body']
    index_messages, nb_messages = _get_index_transitions(email_body)
    structured_historic = [
        {'text': email_body[index_messages[i][1]:index_messages[i+1][0]],
         'meta': email_body[index_messages[i][0]:index_messages[i][1]]
         } for i in range(nb_messages)]

    return structured_historic


def _get_index_transitions(email_body):
    """Returns list of indexes defining the transitions between
    different messages in an email."""
    index = []
    for regex in regex_transition_list:
        for match in re.finditer(regex, email_body):
            idx = (match.start(), match.end())
            index.append(idx)

    index = [(0, 0)] + index
    index = index + [(len(email_body), len(email_body))]
    index = list(set(index))
    index = sorted(index, key=lambda tup: tup[0])
    index = __filter_overlap(index)
    nb_parts = len(index) - 1

    return index, nb_parts


def __filter_overlap(index):
    """Filters indexes in list if they overlap."""
    if len(index) == 2:
        return index
    index_f = []
    i = 0
    j = i + 1
    while j < len(index):
        if index[i][1] > index[j][0]:
            index[i] = (min(index[i][0], index[j][1]),
                        max(index[i][0], index[j][1]))
            j += 1
        else:
            index_f += [index[i]]
            i = j
            j += 1
    index_f += [index[i]]

    return index_f[:i+1]



import pandas as pd
from melusine.prepare_email.mail_segmenting import structure_email

structured_historic = [
    {'text': " \n  \n  \n Bonjours, \n  \n Suite a notre conversation \
téléphonique de Mardi , pourriez vous me dire la \n somme que je vous \
dois afin d'd'être en régularisation . \n  \n Merci bonne journée",
     'meta': ''},
    {'text': " \n Bonjour. \n  \n Merci de bien vouloir prendre connaissance \
du document ci-joint : \n 1 - Relevé d'identité postal MUTUELLE \
(contrats) \n  \n Sentiments mutualistes. \n  \n La Mututelle \n  \n \
La visualisation des fichiers PDF nécessite Adobe Reader. \n  ",
     'meta': ' \n  \n Le mar. 22 mai 2018 à 10:20,  \
<gestionsocietaire@mutuelle.fr> a écrit\xa0:'}]

output = [{'meta': {'date': None, 'from': None, 'to': None},
           'structured_text': {'header': None,
                               'text': [
                                   {'part': ' Bonjours, ',
                                    'tags': 'HELLO'},
                                   {'part': "    Suite a notre conversation \
téléphonique de Mardi , pourriez vous me dire la   somme que je vous dois \
afin d'd'être en régularisation . \n  \n ",
                                    'tags': 'BODY'},
                                   {'part': 'Merci bonne journée',
                                    'tags': 'GREETINGS'}
                                   ]
                               }
           },
          {'meta': {'date': ' mar. 22 mai 2018 à 10:20',
                    'from': '  <gestionsocietaire@mutuelle.fr> ',
                    'to': None},
           'structured_text': {'header': None,
                               'text': [
                                   {'part': ' Bonjour. \n  \n ',
                                    'tags': 'HELLO'},
                                   {'part': "Merci de bien vouloir prendre \
connaissance du document ci-joint :   1 - Relevé d'identité postal MUTUELLE \
(contrats)    ",
                                    'tags': 'BODY'},
                                   {'part': ' Sentiments mutualistes. ',
                                    'tags': 'GREETINGS'},
                                   {'part': '        La Mututelle    ',
                                    'tags': 'BODY'},
                                   {'part': ' La visualisation des fichiers \
PDF nécessite Adobe Reader. \n',
                                    'tags': 'FOOTER'}]}}]


def test_structure_email():
    input_df = pd.DataFrame({
        'structured_historic': [structured_historic]
    })

    output_df = pd.Series([output])

    result = input_df.apply(structure_email, axis=1)
    pd.testing.assert_series_equal(result, output_df)
